from revolutionhtl.in_out import read_tl_digraph
import matplotlib.pyplot as plt
from itertools import combinations, chain
import networkx as nx
from tqdm import tqdm
from statistics import median
from math import sqrt
from collections import Counter

# Reduce giant family
# --------------------

def reduce_graph(giant_0, N_max= 2000):
    undirected_0= nx.Graph()
    undirected_0.add_edges_from(giant_0.edges())
    Q= [(giant_0, undirected_0)]
    res= []
    singles= []
    while len(Q) > 0:
        giant, undirected= Q.pop(0)
        CCs= _search_better(giant, undirected)

        for X in CCs:
            L= len(X)
            if L==1:
                singles+= X
            else:
                if L<=N_max:
                    res+= [X]
                else:
                    Q+= [(nx.induced_subgraph(giant_0, X), nx.induced_subgraph(undirected_0, X))]
                    
    return res, singles

def _compute_partition(undirected):
    CCs= nx.algorithms.community.kernighan_lin_bisection(undirected)
    CCs= list(chain.from_iterable((nx.connected_components(nx.induced_subgraph(undirected, X)) for X in CCs)))
    return CCs

def _search_better(giant, undirected, repetitions= 3):
    CCs= [_compute_partition(undirected) for i in range(repetitions)]
    scores= [_partition_score(X, giant) for X in CCs]
    CCs= CCs[scores.index( max(scores) )]
    return CCs

def _partition_score(CCs, G):
    """
    The biggest, the better
    """
    lens= list(map(len, CCs))
    L= len(lens)
    M= max(lens)

    x1= sum((x>1 for x in lens))/L # proportion of not singletones
    x2= nx.algorithms.community.partition_quality(G, CCs)[0] # proportion of intra-edges
    x3= sum((x/M for x in lens))/L # proportion of size w.r.t. bigest CC

    return sqrt( x1**2 + x2**2 + x3**2 ) # Euclidean norm



"""
Functions to run diamond

Includes standalon usage:
python -m revolutionhtl.diamond -h
"""

from .common_tools import norm_path
from .error import DirectoryExist

from tqdm import tqdm
from itertools import combinations
import os
from subprocess import run

# constants
_default_out_dir= 'tl_project_diamond_all_vs_all'
_default_fasta_ext='.fa'
_default_diamond='diamond'
_diamond_columns= 'qseqid sseqid qlen slen length bitscore evalue'.split()

# Functions
            
def diamond_all_vs_all(fastas_dir,
                       out_dir= _default_out_dir,
                       fasta_ext= _default_fasta_ext,
                       diamond= _default_diamond,
                       quiet= True,
                       evalue= 0.00001, identity= 25, cov= 50,
                       max_target_seqs= 25,
                      ):
    # Normalize and check directories
    fastas_dir, out_dir= _check_path(fastas_dir, out_dir)
    # List fasta files
    is_fasta= lambda x: x.endswith(fasta_ext)
    fasta_files= list(filter(is_fasta, os.listdir(fastas_dir)))
    # Create file names
    species= {file:file.split('.')[0] for file in fasta_files}
    reference= {file:f'{out_dir}{species[file]}' for file in fasta_files}
    f_path= {file:f'{fastas_dir}{file}' for file in fasta_files}
    # Create index
    for file in tqdm(fasta_files, desc= 'Generating index'):
        std_out= diamond_makedb(diamond, f_path[file], reference[file], quiet= quiet)
    # Run pairwise
    n= len(fasta_files)
    for query ,target in tqdm(combinations(fasta_files, 2), total= int(n*(n-1)/2),
                      desc= 'Running alignments'):

        fout= f'{out_dir}{species[query]}.vs.{species[target]}.diamond'
        std_out= diamond_blastp(diamond, f_path[query],
                                reference[target], fout, quiet= quiet,
                                evalue= evalue, identity= identity, cov= cov,
                                max_target_seqs= max_target_seqs
                               )

        fout= f'{out_dir}{species[target]}.vs.{species[query]}.diamond'
        std_out= diamond_blastp(diamond, f_path[target],
                                reference[query], fout, quiet= quiet,
                                evalue= evalue, identity= identity, cov= cov,
                                max_target_seqs= max_target_seqs
                               )
    # End

def diamond_makedb(diamond, fpath, f_reference, quiet= False):
    command= [diamond, 'makedb', '--in',
              fpath, '-d', f_reference]
    if quiet:
        command+= ['--quiet']
    return run(command)

def diamond_blastp(diamond, fpath, f_reference, f_out, quiet= False, evalue= 0.00001, identity= 25, cov= 50, max_target_seqs= 25):
    command= [diamond, 'blastp',
              '-d', f_reference, '-q', fpath, '-o', f_out,
              '-e', evalue, '--id', identity,
              '--query-cover', cov, '--subject-cover', cov, '-k', max_target_seqs,
              '-f', '6'
             ] + _diamond_columns
    command= list(map(str, command))
    if quiet:
        command+= ['--quiet']
    return run(command)

# Auxiliary function
####################

def _check_path(fastas_dir, out_dir):
    # Fastas directory
    fastas_dir= norm_path(fastas_dir)
    if not os.path.exists(fastas_dir):
        raise DirectoryExist(f'There is not directory "{fastas_dir}"')
    # ourput directory
    if out_dir==None:
        out_dir= fastas_dir
    else:
        out_dir= norm_path(out_dir)
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
    # return
    return fastas_dir, out_dir

# Stand-alone usage
###################

if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument('fastas_dir', type= str)
    parser.add_argument('-o', '--out_dir', default= _default_out_dir, type= str)
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-f_ext', '--fasta_ext', default= _default_fasta_ext, type= str)
    parser.add_argument('-d_command', '--diamond', default= _default_diamond, type= str)
    parser.add_argument('-e', '--evalue', default= 0.00001, type= float)
    parser.add_argument('-id', '--identity', default= 25, type= int)
    parser.add_argument('-cov', '--coverture', default= 50, type= int)
    parser.add_argument('-k', '--max_target_seqs', default= 25, type= int)
    args = parser.parse_args()

    # Run
    ######
    hello_diamond= "\n-------------------------------------------------"
    hello_diamond+= "\nRunnng DIAMOND to create REvolutionH-tl input files"
    hello_diamond+= "\n-------------------------------------------------\n"
    print(hello_diamond)
    diamond_all_vs_all(args.fastas_dir,
                       out_dir= args.out_dir,
                       fasta_ext= args.fasta_ext,
                       diamond= args.diamond,
                       quiet= not args.verbose,
                       evalue= args.evalue, identity= args.identity, cov= args.coverture,
                       max_target_seqs= args.max_target_seqs,
                      )
    print("Process finished succesffully\n")



from math import log, ceil
from itertools import chain
import pandas as pd

def LCA(T, nodes):
    if len(nodes)==0:
        raise ValueError(f'Parameter "nodes" must be a non-empty set.')
    idxs= {T.R[x] for x in nodes}
    i,j= get_min_max(idxs)
    #i,j= min(idxs), max(idxs)
    lcaIdx= st_rqm(T.L,i,j,T.st)
    node= T.E[lcaIdx]
    return node

def get_min_max(L):
    L= list(L)
    i=j=L[0]
    for x in L:
        if x<i: i=x
        if x>j: j=x
    return i,j

def get_posible_normalized_blocks(block_size):
    """
    Returns a dictionary where the keys are all the posible
    normalized blocks of size block_size, and the keys are
    the index of the minimum vale.
    """
    res= [(0,)]
    for i in range(block_size-1):
        res= [X+(X[-1]+d,) for X in res for d in [1,-1]]
    #res= {X: X.index(min(X)) for X in res}
    return res

def compute_perfect_size(L):
    k= ceil(log(len(L), 2))
    n= 2**k
    return n, k


def adjust_array_size(L, n):
    L= L.copy()
    n0= len(L)
    l= L[-1]
    for i in range(n-n0):
        L= L+[l]
    return L

def get_blocks(L, block_size, n_blocks):
    return (tuple(L[block_size*i : block_size*(i+1)])
            for i in range(n_blocks))

def normalize_block(L):
    y0= L[0]
    return tuple((y-y0 for y in L))

def normalize_array(L, block_size, n_blocks):
    return list(chain.from_iterable((normalize_block(L0)
                                     for L0 in get_blocks(L, block_size, n_blocks))))

def aux_arrays(L, block_size, n_blocks, norm_block_2_min_idx):
    A=[]
    B=[]
    for L0 in get_blocks(L, block_size, n_blocks):
        b= norm_block_2_min_idx.get(normalize_block(L0),
                                    L0.index(min(L0)))
        A+= [L0[b]]
        B+= [b]
    return A,B

# ST-RMQ
def st_rqm(L,i,j,st):
    if i==j:
        return i
    i,j= sorted((i,j))
    k= int( log(j-i+1, 2) )
    a= int( st.loc[i, k-1] )
    b= int( st.loc[j-(2**k)+1, k-1] )

    return choose_argmin(L, a, b)

# Compute sparse table
######################
def get_range(n0, j):
    w_size= 2**(j)
    return range(n0 - w_size +1)

def choose_argmin(L,i,j):
    if L[i] <= L[j] :
        return i
    return j

def compute_sparce_table(L):
    n0,k= compute_perfect_size(L)
    L= adjust_array_size(L, n0)

    M= [[choose_argmin(L, i, i+1) for i in get_range(n0, 1)]]

    for j in range(2, k+1):
        dM= []
        R= get_range(n0, j)
        for i in R:
            b0= i+(2**(j-1))
            a= M[ j-2 ][ i ]
            b= M[ j-2 ][ b0 ]
            dM+= [choose_argmin(L, a, b)]
        M+= [dM]

    return pd.DataFrame(M).T




from .nxTree import is_leaf, get_dad
from .lca_rmq import LCA
from .triplets import get_triplets, _get_triples_from_root
from .prune import prune_losses
from .nhxx_tools import get_nhx

import networkx as nx
import pandas as pd
from itertools import chain

from tqdm import tqdm
tqdm.pandas()

# Extra variables for consistency testing:
# - ind_leafs
# - Ts_triples

def assign_id(T, idAttr= 'node_id'):
    for x in T:
        T.nodes[x][idAttr]= x

def _try_recon(Tg, Ts, Ts_triples, root, gene_attr, species_attr, event_attr, fix_inconsistencies):

    X= reconciliate(Tg, Ts, Ts_triples, root, gene_attr, species_attr, event_attr, fix_inconsistencies)

    Tr, mu, epsilon, flip_nodes= X
    return Tr, mu, epsilon, flip_nodes

def reconciliate_many(gTrees,
                      Ts,
                      root= 1,
                      gene_attr= 'accession',
                      species_attr= 'species',
                      event_attr= 'accession',
                      fix_inconsistencies= True,
                      orthogroup_column= 'OG'):
    for y in Ts:
        if len(Ts[y])>0:
            Ts.nodes[y][event_attr]= 'S'

    Ts_triples= set(get_triplets(Ts, event=event_attr, color=species_attr, root_event='S'))

    # Reconciliate trees
    #   Note:
    #   Each element X of the list bellow is a tuple
    #   containing: X[0]-> Tr, X[1]-> mu, X[2]-> epsilon, X[3]-> flip_nodes
    recons= [(idx,) + _try_recon(Tg,
                                 Ts,
                                 Ts_triples,
                                 root,
                                 gene_attr,
                                 species_attr,
                                 event_attr,
                                 fix_inconsistencies)
             for Tg,idx in tqdm(zip(gTrees, gTrees.index),total= gTrees.shape[0])]

    df_recs= pd.DataFrame(recons, columns= [orthogroup_column,
                                            'reconciliated_tree',
                                            'reconciliation_map',
                                            'equivalence_map',
                                            'flipped_nodes'],
                         )

    return df_recs

def recon_table_to_str(Ts, df_recs, orthogroup_column= 'OG'):
    # Assign node-IDs for the newick output
    assign_id(Ts)
    df_recs.reconciliated_tree.apply(assign_id)

    df_r= pd.DataFrame(dict(tree= df_recs.reconciliated_tree.apply( lambda x: get_nhx(x, name_attr= 'accession') ),
                            reconciliation_map= df_recs.apply(lambda row: recon_2_text(row.reconciliation_map,
                                                                                       row.reconciliated_tree,
                                                                                       Ts
                                                                                      ), axis= 1),
                            flipped_nodes= df_recs.apply(lambda row: flipped_to_str(row.flipped_nodes,
                                                                                    row.reconciliated_tree
                                                                                   ) , axis= 1)
                           ))
    df_r[orthogroup_column]= df_recs[orthogroup_column]

    return df_r[[orthogroup_column, 'tree', 'reconciliation_map', 'flipped_nodes']]

def flipped_to_str(flipped, Tr, idAttr= 'node_id'):
    return ','.join( (str(get_node_id(Tr, x, idAttr)) for x in flipped) )

def get_node_id(T, x, idAttr= 'node_id'):
    return T.nodes[x][idAttr]

def recon_2_text(mu, T0, T1, idAttr= 'node_id'):
    G= lambda X: f'{ get_node_id(T0, X[0], idAttr) }:{ get_node_id(T1, X[1], idAttr) }'
    return ','.join( (G(X) for X in mu.items()) )

def reconciliate(Tg, Ts, Ts_triples, root= None, gene_attr= 'accession', species_attr= 'species', event_attr= 'event', fix_inconsistencies= True):
    Tr= nx.DiGraph()
    mu= dict()
    epsilon= dict()
    x_node= 0
    flip_nodes= []
    ind_leafs= {} # Induced species for nodes into the input tree
    if root == None:
        root= Tg.root

    for u_node in nx.dfs_postorder_nodes(Tg, source= root):
        if is_leaf(Tg, u_node):
            rho_x, x_node= _map_leaf(Tg, u_node, Tr, x_node, Ts, mu, species_attr, gene_attr)
            ind_leafs[u_node]= { Tg.nodes[u_node][species_attr] }
        else:
            lamb_x= [epsilon[v] for v in Tg[u_node]]
            lamb_y= {mu[x] for x in lamb_x}
            y= LCA(Ts, lamb_y)
            rho_x, x_node= _rarefy(Ts, Tr, Tg, ind_leafs, Ts_triples, y, u_node, event_attr, lamb_y, fix_inconsistencies, mu, lamb_x, x_node, flip_nodes)
            ind_leafs[u_node]= set(chain.from_iterable((ind_leafs[u_child] for u_child in Tg[u_node])))
        epsilon[u_node]= rho_x

    Tr.root= epsilon[root]

    return Tr, mu, epsilon, flip_nodes


################################################################################
################################################################################
def _map_leaf(Tg, u_node, Tr, x_node, Ts, mu, species_attr, gene_attr):
    rho_x= x_node
    x_node+= 1
    species= Tg.nodes[u_node][species_attr]
    gene= Tg.nodes[u_node][gene_attr]
    Tr.add_node(rho_x, **{species_attr:species, gene_attr:gene})
    sleafs= [x for x in Ts if len(Ts[x])==0 and Ts.nodes[x][species_attr] == species]
    ks= len(sleafs)
    if ks!=1:
        raise ValueError(f'There should be exactly one leaf in the species tree with species "{species}", but there are {ks}')
    mu[rho_x]= sleafs[0]
    return rho_x, x_node

def _rarefy(Ts, Tr, Tg, ind_leafs, Ts_triples, y, u_node, event_attr, lamb_y, fix_inconsistencies, mu, lamb_x, x_node, flip_nodes):
    if Tg.nodes[u_node][event_attr]=='S':
        #<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< In the future we should use the "indirect" function (now commented)
        #rho_x, x_node= _in_direct_triple_testing(u_node, y, Tr, Ts, mu, lamb_x, lamb_y, x_node, event_attr, flip_nodes)
        rho_x, x_node= _direct_triple_testing(u_node, y, Tr, Tg, Ts, ind_leafs, Ts_triples, mu, lamb_x, lamb_y, x_node, event_attr, flip_nodes)
    elif Tg.nodes[u_node][event_attr]=='D':
        rho_x, x_node= _rarefy_duplication(u_node, y, Tr, Ts, mu, lamb_x, x_node, event_attr)
    else:
        raise ValueError(f'Wrong symbol "{Tg.nodes[u_node][event_attr]}" in the gene tree.')
    return rho_x, x_node

def _compute_xi(rho_y, Y_0, Ts):
    Xi= {rho_y}
    Y_nodes= {get_dad(Ts, y) for y in Y_0 if rho_y not in (y, get_dad(Ts, y))}
    while len(Y_nodes)>0:
        y_node= Y_nodes.pop()
        if y_node in Y_0:
            return set()
        if y_node not in Xi:
            Xi.add(y_node)
            Y_nodes.add(get_dad(Ts, y_node))
    return Xi

def _rarefy_speciation(Xi, X_0, Y_0, Ts, Tr, mu, y_0, x_node, event_attr):
    """
    We assume Y_0 < y_0
    """
    Ts_p= nx.induced_subgraph(Ts, Xi.union(Y_0))
    delta= {mu[x]:x for x in X_0}
    for y_node in nx.dfs_postorder_nodes(Ts_p, source= y_0):
        children_Ts_p= list(Ts_p[y_node])

        if len(children_Ts_p) > 0:
            # Create the node
            x_iter_p= x_node
            x_node+= 1
            Tr.add_node(x_iter_p, **{event_attr:'S'})
            mu[x_iter_p]= y_node
            delta[y_node]= x_iter_p
            # Identify the map of the children of x_iter
            children_Tr= map(lambda y: delta[y], children_Ts_p)
            # Connect the childres via inverse map delta
            for x_pp in children_Tr:
                Tr.add_edge(x_iter_p, x_pp)
            # Create missed child-withness
            for y_pp in set(Ts[y_node]) .difference( children_Ts_p ) :
                Tr.add_node(x_node, **{event_attr:'X'})
                Tr.add_edge(x_iter_p, x_node)
                mu[x_node]= y_pp
                x_node+= 1
    return x_iter_p, x_node

def _rarefy_duplication(u_node, rho_y, Tr, Ts, mu, X_nodes, x_node, event_attr):
    # Create the duplication
    rho_x= x_node
    x_node+= 1
    Tr.add_node(rho_x, **{event_attr:'D'})
    mu[rho_x]= rho_y
    # Create descendant linages
    for x_node_p in X_nodes:
        if rho_y==mu[x_node_p]:
            Tr.add_edge(rho_x, x_node_p)
        else:
            lamb_x= {x_node_p}
            lamb_y= {mu[x_node_p]}
            Xi= _compute_xi(rho_y, lamb_y, Ts)
            rho_x_p, x_node= _rarefy_speciation(Xi, lamb_x, lamb_y, Ts, Tr, mu, rho_y, x_node, event_attr)
            Tr.add_edge(rho_x, rho_x_p)
    return rho_x, x_node

def _direct_triple_testing(u_node, y, Tr, Tg, Ts, ind_leafs,
                           Ts_triples, mu, lamb_x, lamb_y,
                           x_node, event_attr, flip_nodes):
    Tg_color_triples= set(_get_triples_from_root(Tg, u_node, ind_leafs))
    FP= Tg_color_triples - Ts_triples
    if len(FP)>0:
        rho_x, x_node= _rarefy_duplication(u_node, y, Tr, Ts, mu,
                                           lamb_x, x_node, event_attr)
        flip_nodes+= [rho_x]#[u_node]
    else:
        Xi= _compute_xi(y, lamb_y, Ts)
        rho_x, x_node= _rarefy_speciation(Xi, lamb_x, lamb_y, Ts, Tr, mu,
                                          y, x_node, event_attr)
    return rho_x, x_node

def _in_direct_triple_testing(u_node, y, Tr, Ts, mu, lamb_x, lamb_y, x_node, event_attr, flip_nodes):
    if len(lamb_y) < len(lamb_x): # There are nodes sharing the same map
        rho_x, x_node= _rarefy_duplication(u_node, y, Tr, Ts, mu, lamb_x, x_node, event_attr)
        flip_nodes+= [rho_x]#[u_node]
    if y in lamb_y: # Time insocnsistency
        rho_x, x_node= _rarefy_duplication(u_node, y, Tr, Ts, mu, lamb_x, x_node, event_attr)
        flip_nodes+= [rho_x]#[u_node]
    else:
        Xi= _compute_xi(y, lamb_y, Ts)
        if len(Xi)==0: # All refinements are non-consistent
            rho_x, x_node= _rarefy_duplication(u_node, y, Tr, Ts, mu, lamb_x, x_node, event_attr)
            flip_nodes+= [rho_x]#[u_node]
        else:
            rho_x, x_node= _rarefy_speciation(Xi, lamb_x, lamb_y, Ts, Tr, mu, y, x_node, event_attr)
    return rho_x, x_node



import networkx as nx
import pandas as pd
from itertools import chain

def cBMG_not_cograph():
    df= pd.read_csv('../examples/cBMG_not_cograph.txt', sep= '\t')
    nodes= set( chain.from_iterable( df.values ) )
    IDXs= {x:i for i,x in enumerate(nodes)}

    G= nx.DiGraph()
    F= lambda x: x[:-1]
    for x in nodes:
        G.add_node(IDXs[x], gene= x, species= F(x))
    G.add_edges_from(map(lambda X: list(map(lambda x: IDXs[x], X)) , df.values))
    return G

def BMG_2colored():
    """
    Two colores best match graph
    """
    G= nx.DiGraph()
    blue_nodes= ['a1', 'a2', 'a3', 'a4', 'a5', 'a6']
    red_nodes= ['b1', 'b2', 'b3', 'b4']
    nodes= blue_nodes + red_nodes
    nLabel= {x:i for i,x in enumerate(nodes)}
    #node_attrs= [(x, {'color': 'red'}) for x in blue_nodes] + [(x, {'color': 'blue'}) for x in red_nodes]
    node_attrs=  [(nLabel[x], {'gene': x, 'species': 'red'}) for x in blue_nodes]
    node_attrs+= [(nLabel[x], {'gene': x, 'species': 'blue'}) for x in red_nodes]

    G.add_nodes_from(node_attrs)

    edges= [('a1', 'b1'),
            ('a1', 'b2'),
            ('a1', 'b3'),
            ('a1', 'b4'),
            ('a2', 'b2'),
            ('a3', 'b3'),
            ('a4', 'b4'),
            ('a5', 'b2'),
            ('a5', 'b3'),
            ('a5', 'b4'),
            ('a6', 'b3'),
            ('a6', 'b4'),
            ('b1', 'a1'),
            ('b1', 'a2'),
            ('b1', 'a3'),
            ('b1', 'a4'),
            ('b1', 'a5'),
            ('b1', 'a6'),
            ('b2', 'a2'),
            ('b3', 'a3'),
            ('b4', 'a4'),
           ]

    edges= [(nLabel[ed[0]], nLabel[ed[1]]) for ed in edges]

    G.add_edges_from(edges)
    return G



import networkx as nx
from itertools import permutations
#from build_st_n import BUILDst_N
from . import nxTree as nxt
from .error import TreeDontExplainGraph

def LRT_from_cBMG(G, speciesAttr, geneAttr):
    triplets= _get_informative_triplets(G, geneAttr, speciesAttr)
    T= BUILDst_N( triplets, geneAttr, speciesAttr)
    G1= get_cBMG_of_T(T, speciesAttr, geneAttr)
    if not _are_equal_graphs(G, G1, geneAttr):
        raise TreeDontExplainGraph('The graph is not a cBMG')
    return T

def get_cBMG_of_T(T, color_attr, geneAttr):
    G1= nx.DiGraph()
    G1.add_edges_from( from_tree_2_cBMs(T, color_attr, geneAttr) )
    return G1

def from_tree_2_cBMs(T, color_attr, geneAttr):
    best_matches= []
    color_2_leafs= nxt.get_color2leafs(T, color_attr)
    for color0,color1 in permutations(color_2_leafs, 2):
        for x in color_2_leafs[color0]:
            best_matches+= [[x,y] for y in _get_best_matches_of_x(x, color_2_leafs[color1], T)]
    F= lambda x: T.nodes[x][geneAttr]
    return [list(map(F, X)) for X in best_matches]

def _are_equal_graphs(G, G1, geneAttr):
    are_equal= True
    for x in G:
        #are_equal= set(G[x]) == set(G1[x])
        x1= G.nodes[x][geneAttr]
        are_equal= {G.nodes[y][geneAttr] for y in G[x]} == set(G1[x1])
        if not are_equal:
            break
    return are_equal

def _get_informative_triplets(G, geneAttr, speciesAttr):
    """
    ...
    """
    DD= _get_color_2_nodes(G, speciesAttr)
    ret= []
    for color0, color1 in permutations(DD, 2):
        others= set(DD[color1])
        for node in DD[color0]:
            ret+= _get_node_inf_triplets(G, node, others, [color0, color1], geneAttr, speciesAttr)
    return ret

def _get_node_inf_triplets(G, node, others, colors, geneAttr, speciesAttr):
    """
    Se harán todas las tripletas de tipo 'node x | y'
    Donde el color de 'x' y 'y' son el mismo, pero diferente del color de 'node'.
    Además 'node -> x' en el grafo, pero no pasa que 'node -> y'

    Input:
    - G
    - node
    - others: Todos los nodos de un mismo color
    """
    neighbors= set(G[node])
    no_neighbors= others - neighbors
    return [_create_triplet(node, x, y, G, geneAttr, speciesAttr)
            for x in neighbors for y in no_neighbors]

def _create_triplet(node, x, y, G, geneAttr, speciesAttr):############ Revisar bien lo que debería hacer esta función y normalizar
    """
    The triplet is 'node x | y'
    """
    T= nx.DiGraph()
    T.root= 0
    T.add_nodes_from([0,1], species= None, gene= None)
    T.add_node(2, gene= G.nodes[y][geneAttr],
               species= G.nodes[y][speciesAttr])
    T.add_node(3, gene= G.nodes[node][geneAttr],
               species= G.nodes[node][speciesAttr])
    T.add_node(4, gene= G.nodes[x][geneAttr],
               species= G.nodes[x][speciesAttr])
    T.add_edges_from(((0,1),(0,2),(1,3),(1,4)))
    return T

def _get_color_2_nodes(G, colorAttr):
    DD= {}
    for node in G:
        color= G.nodes[node][colorAttr]
        DD[ color ]= DD.get(color, list()) + [node]
    return DD

def _compute_lca_hash(x, nodes, T):
    Y_x_t= {}
    for y in nodes:
        v= nxt.LCA(T, [x,y])
        Y_x_t[v]= Y_x_t.get(v, []) + [y]
    return Y_x_t

def _get_best_matches_of_x(x, nodes, T):
    Y_x_t= _compute_lca_hash(x, nodes, T)
    Y= list(Y_x_t)
    best= Y[0]
    for y in Y[1:]:
        if nxt.LCA(T, [best, y]) == best :
            best= y
    return Y_x_t[ best ]



class InconsistentTrees(Exception):
    """Raised when input trees are not consistent when they should"""
    pass

class TreeDontExplainGraph(Exception):
    """Raised when a given tree doesn't explain a given graph"""
    pass

class NotProperlyColoredDigraph(Exception):
    """Raised when a suposed cBMG have different color sets in its connected components"""
    pass

class NotSinkFreeDigraph(Exception):
    """Raised when a suposed cBMG is not sink free"""
    pass

class MissedData(Exception):
    """Raised when trying to run a step in revolutionhtl.__main__ without providing the proper data"""
    pass

class ParameterError(Exception):
    """Raised when trying to run revolutionhtl.__main__ with a wrong parameter"""
    pass

# General errors
################
class DirectoryExist(Exception):
    """Raised when trying create a directory that already exists"""
    pass



from .lca_rmq import LCA, compute_sparce_table

from itertools import product, takewhile, combinations, chain
import pandas as pd
import numpy as np
from collections.abc import Iterable
import networkx as nx

__forbidden_attributes_phylo= ['_color', 'clades']


#> Define some functions
########################

# Para árboles

def get_dad(tree, n):
    in_edgs= list(tree.in_edges(n))
    if len(in_edgs)==0:
        if n in tree:
            raise ValueError('Node "{}" has no dad'.format(n))
        else:
            raise ValueError('Node "{}" not in tree'.format(n))
    return in_edgs[0][0]

def get_path(tree, x0, x1):
    P= [x1]
    while( x1 != x0 ):
        x1= get_dad(tree, x1)
        P+= [x1]
    return P[::-1]

def are_the_same(X):
    x0= X[0]
    F= lambda x: x==x0
    return all(map(F, X))

def LCA_naive(tree, nodes):
    if len(nodes)==0:
        raise ValueError(f'Parameter "nodes" must be a non-empty set.')
    P= [get_path(tree, tree.root, n) for n in nodes]
    L= min(map(len, P))
    P= [X[:L] for X in P]
    M= len(nodes)
    #X= list( takewhile(lambda x: are_the_same(P[:,x]) , range(L)) )[-1]    
    X= list( takewhile(lambda x: are_the_same( [P[i][x] for i in range(M)] ) , range(L)) )[-1]    
    return P[0][X]


def LCA_color(tree, colors, color_attr= 'color'):
    if len(colors)==0:
        raise ValueError(f'Parameter "colors" must be a non-empty set.')
    nodes= {x for x in tree if tree.nodes[x][color_attr] in colors}
    return LCA(tree, nodes)

def set_sparce_matrix(T):
    # Add attributes for rmq-lca
    E, L, R= euler_tour(T)
    T.E= E
    T.R= R
    T.L= L
    T.st= st= compute_sparce_table(L)
    return None

def induced_leafs(tree, n):
    Q= [n]
    L= []
    while len(Q) > 0:
        n= Q.pop()
        sons= list(tree[n])
        if len(sons) == 0:
            L+= [n]
        else:
            Q+= sons
    return L

def get_inner_nodes(tree):
    return set(tree) - set(induced_leafs(tree, tree.root))

def is_leaf(tree, node):
    return len(tree[node]) == 0

def get_color2leafs(tree, color_attr= 'color'):
    leafs= induced_leafs(tree, tree.root)
    colors= set((tree.nodes[leaf][color_attr] for leaf in leafs))
    color_2_leaf= {color:set() for color in colors}
    for leaf in leafs:
        color_2_leaf[ tree.nodes[leaf][color_attr] ].add(leaf)
    return color_2_leaf

def get_leaf2color(T, color_attr= 'color'):
    leaves= filter(lambda x: is_leaf(T,x), T)
    return {x:T.nodes[x][color_attr] for x in leaves}

def get_symbolic_relations(T, sym, nAttr, forbidden_nodes= [], leaf_attr= None, forbidden_leaf_symbol= 'X'):
    if leaf_attr==None:
        F= lambda X: X
    else:
        F= lambda X: [T.nodes[x_l][leaf_attr]
                      for x_l in X
                      if T.nodes[x_l][leaf_attr] !=forbidden_leaf_symbol
                     ]
    edges= []
    for node in T:
        if T.nodes[node][nAttr]==sym and node not in forbidden_nodes:
            for x0,x1 in combinations(T[node], 2):
                edges+= list(product(F(induced_leafs(T, x0)),
                                     F(induced_leafs(T, x1)),
                                    ))
    return edges

def copy_tree(tree):
    root= tree.root
    tree= tree.copy()
    tree.root= root
    return tree

def subtree(T, root):
    nodes= BFS_graph(T, root)
    T1= nx.induced_subgraph(T, nodes)
    T1.root= root
    return T1

# Edge compacting
def compact_nondiscriminant_edges(T, attr= 'label', attr_val= None, root= None):
    # Set parameters
    if attr_val == None:
        are_equal= lambda x0,x1: T.nodes[x0][attr] == T.nodes[x1][attr]
    else:
        are_equal= lambda x0,x1: T.nodes[x0][attr] == T.nodes[x1][attr] == attr_val
    if root==None:
        root= T.root
    # Compact edges
    edges= list((x0,x1) for x0 in nx.dfs_postorder_nodes(T, root) for x1 in T[x0])
    for x0,x1 in edges:
        if are_equal(x0,x1):
            compact_edge(T, x0, x1, update_lca= False)
    # Update LCA matrix
    set_sparce_matrix(T)
    return None

def compact_edge(T, x0, x1, delete= 'lower', update_lca= True):
    if delete=='lower':
        Y= set(T[x1])
        T.remove_node(x1)
        for y in Y:
            T.add_edge(x0,y)
    elif delete=='upper':
        Y= set(T[x0]) - {x1}
        dad= get_dad(T, x0)
        T.remove_node(x0)
        T.add_edge(dad,x1)
        for y in Y:
            T.add_edge(x1,y)
    else:
        ValueError('delete parameter must be one of ["lower", "upper"]')
    if update_lca:
        set_sparce_matrix(T)
    return None

# Prunnng
def prune_losses(T, label='label', loss='X', root= None):
    # Set parameters
    if root==None:
        root= T.root
    # Prune losses
    node_2_children= dict()
    indiced_L= dict()
    # Do a DFS travel: save children in preorder, prune nodes in posorder
    stack= [root]
    while (len(stack)>0):
        peak= stack[-1]
        if peak in node_2_children:
            # Prune if necessary
            _check_if_prune(T, peak, node_2_children[peak], indiced_L, label, loss)
            # Update stack
            stack.pop(-1)
        else:
            # Update children dictionary and stack
            node_2_children[peak]= set(T[peak])
            stack+= list(T[peak])
    # Update LCA matrix
    set_sparce_matrix(T)
    return None

def _check_if_prune(T, x_node, children, indiced_L, label, loss):
    # compute induced leafs
    if is_leaf(T, x_node):
        indiced_L[x_node]= {T.nodes[x_node][label]}
    else:
        indiced_L[x_node]= set(chain.from_iterable((indiced_L[x] for x in children)))
    # Chech if prune
    if indiced_L[x_node] == {loss}:
        # All descendants are loss, thus delete node
        T.remove_node(x_node)
    elif len(T[x_node])==1:
        # There is only one child, thus compact node
        x1= list(T[x_node])[0]
        compact_edge(T, x_node, x1, delete= 'upper', update_lca= False)
    return None


# Leafs overlap
def are_overlapping(set_list):
    ret= False
    for s0,s1 in combinations(set_list, 2):
        if not s0.isdisjoint(s1):
            ret= True
            break
    return ret

def leafs_overlap(tree, node):
    if not is_leaf(tree, node):
        return are_overlapping([induced_colors(tree, child)
                                for child in tree[node]])
    else:
        raise ValueError(f'leafs overlap not defined for a leaf ({node})')


def assign_symbolic_date(tree):
    for node in tree:#BFS_graph(tree, tree.root):
        if not is_leaf(tree, node):
            if leafs_overlap(tree, node):
                tree.nodes[node]['color']= 'D'
            else:
                tree.nodes[node]['color']= 'S'

# Pasar a otros tipos de datos

def get_newick(T):
    newick= {}
    for node in nx.dfs_postorder_nodes(T):
        if len(T[node])==0 :
            newick[ node ]= 'spec' + node#nxTree.nodes[node]['label']
        else:
            newick[ node ]= '(' + ','.join([ newick[son] for son in T[node] ]) + ')'
    return newick[node]+';'

def __getNAme(node, idx, tree, label):
    attributes= {a:b for a,b in node.__dict__.items()
                 if b != None
                 and a not in __forbidden_attributes_phylo
                }
    n_label= attributes.get(label, None)
    if n_label==None:
        return idx, attributes
    if n_label in tree:
        raise ValueError('The label "{}" is not unique for every node'.format(n_label))
    return n_label, attributes

def make_directed(X):
    """
    Transforma un árbol con estrucutura de dato nx.Graph  a un objeto de networkx.Digraph sin ningun elemento de biopython
    """
    tree= nx.DiGraph()
    L= []
    for x0 in BFS_graph(X.nx_tree, X.bio_tree.root):
        L+= [x0]
        for x1 in X.nx_tree[x0]:
            if x1 not in L:
                w= X.nx_tree[x0][x1]['weight']
                tree.add_edge(x0, x1, weight= w)

    tree.root= X.bio_tree.root
    return tree


# Recorridos

def BFS_graph(tree, root):
    """
    Returns a list of nodes visited in a BFS travel

    tree: nx.Graph
    root: node in nx.Graph.nodes
    """
    Q= [root]
    L= []
    while len(Q) > 0:
        x= Q.pop(0)
        L+= [x]
        Q+= [neighbor for neighbor in tree[x]
             if neighbor not in L]
    return L

def euler_tour(T):
    """
    input: nxTree
    output:
      E: Euler tour nodes
      L: Levels
      R: Representative index of nodes in E
    """
    stack= [T.root]
    children= {}

    E= []
    R= {}
    L= []
    idx= 0
    lev= 0

    while len(stack)>0:
        node= stack[-1]
        E+= [node]
        L+= [lev]

        if node not in children:
            children[node]= list(T[node])
            R[node]= idx

        if len(children[node])==0:
            stack.pop(-1)
            lev-= 1
        else:
            stack+= [children[node].pop(0)]
            lev+= 1

        idx+= 1

    return E, L, R

# auxiliares

def is_superColor(color):
    return isinstance(color, Iterable) and not type(color)==str

def correct_colors(tree):
    for n in tree:
        colorts= tree.nodes[n]['color']
        #if( type(colorts)==list ):
        #if isinstance(colorts, Iterable) :
        if is_superColor(colorts):
            tree.nodes[n]['color']= list(map(str, colorts))
        else:
            tree.nodes[n]['color']= str(colorts)
    tree.root= '0'
    return tree

def induced_colors(tree, node, color_attr= 'color'):
    sLeafs= set()
    for x in induced_leafs(tree, node):
        xcol= tree.nodes[x][color_attr]
        if is_superColor(xcol):
            sLeafs.update(xcol)
        else:
            sLeafs.add(xcol)
    return sLeafs





import networkx as nx
from itertools import chain, product
from .nxTree import set_sparce_matrix

_name_attr= 'label'

def tralda_to_nxTree(tralda_tree, name_attr= 'accession'):
    newick= tralda_tree.to_newick()
    nxT= read_nhxx(newick, name_attr= name_attr)
    return nxT

def _add_node(T, node, dad):
    T.add_node(node)
    T.add_edge(dad, node)

def _add_node_info(T,
                   node,
                   cache_label,
                   cache_weigth,
                   cahce_dict,
                   name_attr= _name_attr,
                   branch_length_attr= 'length',
                  ):
    if cache_label != None:
        T.nodes[node][name_attr]= cache_label
    else:
        T.nodes[node][name_attr]= node

    if cache_weigth != None:
        T.nodes[node][branch_length_attr]= float(cache_weigth)

    T.nodes[node].update(cahce_dict)

def _update_cache_dict(cache_key, cache_value, cahce_dict, attrs_type, use_nhx_flag, add_attrs):
    if add_attrs:
        typer= attrs_type.get(cache_key, lambda x: x)
        cahce_dict[cache_key]= typer(cache_value)

def ask_attrs(add_attrs, use_nhx_flag, cache_key):
    if use_nhx_flag:
        if cache_key=='&&NHX':
            add_attrs= True
    else:
        add_attrs= True
    return add_attrs

def read_nhxx(nhx,
              name_attr= _name_attr,
              attr_sep= ';',
              attrs_type= {},
              use_nhx_flag= False,
              fill_missed_attrs= True,
             ):
    # Init tree
    T= nx.DiGraph()
    T.add_node(0, **{name_attr: 0})
    T.root= 0

    # Auxiliar variables
    current= 0    # Current node
    idx= 1        # Index for new nodes
    S= [0]        # Stack of visited inner nodes
    last_controller= None

    reading_label= False
    cache_label= None
    reading_weight= False
    cache_weigth= None
    reading_comment= False
    cache_key= None
    cache_value= None
    cahce_dict= {}
    clasp_count= 0 # Cuenta si hay caracteres '[' en el comentario, todos deben ser cerrados.
    is_key= False
    is_value= False
    add_attrs= False
    end_comment= None

    for x in nhx:
        if reading_comment:
            if x=='[':
                clasp_count+= 1
            elif x==']':
                if clasp_count==0:
                    add_attrs= ask_attrs(add_attrs, use_nhx_flag, cache_key)
                    _update_cache_dict(cache_key, cache_value, cahce_dict, attrs_type, use_nhx_flag, add_attrs)
                    reading_comment= False
                    is_key= is_value= False
                    cache_key= cache_value= None
                    add_attrs= False
                    end_comment= True
                else:
                    clasp_count-= 1
            if x=='=':
                is_key= False
                is_value= True
                cache_value= ''
            elif x==attr_sep:
                add_attrs= ask_attrs(add_attrs, use_nhx_flag, cache_key)
                _update_cache_dict(cache_key, cache_value, cahce_dict, attrs_type, use_nhx_flag, add_attrs)
                is_key= is_value= False
                cache_key= cache_value= None
            elif is_key:
                cache_key+= x
            elif is_value:
                cache_value+= x
            elif not end_comment:
                is_key= True
                is_value= False
                cache_key= x
        elif x=='(':
            last_controller= '('
            # Create node
            _add_node(T, idx, current)
            S+= [idx]
            current= idx
            idx+= 1
        elif x==',':
            last_controller= ','
            # End of node
            _add_node_info(T, current, cache_label, cache_weigth, cahce_dict, name_attr= name_attr)
            cache_label= cache_weigth= None
            cahce_dict= {}
            reading_label= reading_weight= reading_comment= False
            current= S[-1]
        elif x==')':
            last_controller= ')'
            # End of node
            _add_node_info(T, current, cache_label, cache_weigth, cahce_dict, name_attr= name_attr)
            cache_label= cache_weigth= None
            cahce_dict= {}
            reading_label= reading_weight= reading_comment= False
            # Exit node
            current= S.pop()
        elif x=='[':
            last_controller= '['
            reading_comment= True
            end_comment= False
        elif x==':':
            last_controller= ':'
            reading_weight= True
            reading_label= False
            cache_weigth= ''
        elif x==';':
            last_controller= ';'
            # End of node
            _add_node_info(T, current, cache_label, cache_weigth, cahce_dict, name_attr= name_attr)
            # Pop last element of stack
            S.pop()
            break
        else:
            # Reading node name or weight
            if reading_label:
                cache_label+= x
            elif reading_weight:
                cache_weigth+= x
            else:
                reading_label= True
                cache_label= x
                if last_controller!=')':
                    _add_node(T, idx, current)
                    current= idx
                    idx+= 1
    if len(S)>0:
        msg= f'";" found before closing all the parenthesis: {S}'
        raise ValueError(msg)

    if fill_missed_attrs:
        all_attrs= set(chain.from_iterable( dict(T.nodes(data= True)).values() ))
        for x,attr in product(T, all_attrs):
            if attr not in T.nodes[x]:
                T.nodes[x][attr]= None

    # Add attributes for rmq-lca
    set_sparce_matrix(T)

    return T

def get_nhx(T,
            root= None,
            name_attr= None,
            attr_sep= ';',
            ignore_attrs= [],
            ignore_inner_name= False,
            include_none= False,
            use_nhx_flag=False,
           ):
    if not root:
        root= T.root

    newick= {}
    for node in nx.dfs_postorder_nodes(T, source= root):
        if len(T[node])==0 :
            nwk_n= str(T.nodes[node].get(name_attr, ''))
        else:
            nwk_n= '(' + ','.join([ newick[son] for son in T[node] ]) + ')'
            if not ignore_inner_name:
                nwk_n+= str(T.nodes[node].get(name_attr, ''))

        c_attrs= set((at for at in T.nodes[node] if (include_none or T.nodes[node][at]!=None)))
        c_attrs-= set(ignore_attrs+[name_attr])
        if len(c_attrs) > 0:
            nwk_n+= '[' + attr_sep.join((f'{x}={T.nodes[node][x]}' for x in c_attrs)) +']'

        newick[ node ]= nwk_n
    return newick[root]+';'



from .nxTree import compact_edge, get_dad, set_sparce_matrix,get_leaf2color, is_leaf
from .nhxx_tools import read_nhxx, get_nhx
from .triplets import get_triplets

import time
import networkx as nx
from networkx import dfs_postorder_nodes
import pandas as pd
from itertools import chain
from tqdm import tqdm
tqdm.pandas()

def correct_tree(Tg, Rs, root= 1,
                 label_attr= 'label',
                 species_attr= 'species',
                 event_attr= 'event',
                 algorithm= 'prune_L',
                 force_phylogeny= True,
                 update_lca= True, 
                 inplace= True,
                ):

    # Inconsistent gene triples
    leaf_2_color= {Tg.nodes[x][label_attr] : Tg.nodes[x][species_attr]
                   for x in Tg if is_leaf(Tg, x)}
    F= lambda x: leaf_2_color[x]
    g2c_triple= lambda triple: tuple(sorted(map(F, triple[:2]))) + (F(triple[-1]),)

    R_I= set()
    R_C= set()
    for rg in get_triplets(Tg, event=event_attr, color=label_attr, root_event='S'):
        if g2c_triple(rg) in Rs:
            R_C.add(rg)
        else:
            R_I.add(rg)

    # From label to node
    leaf_2_node= {Tg.nodes[x][label_attr] : x
                  for x in Tg if is_leaf(Tg, x)}
    F= lambda triple: tuple(leaf_2_node[x] for x in triple)
    R_C= set(map(F, R_C))
    R_I= set(map(F, R_I))

    # Prune
    if algorithm=='prune_R':
        ret= prune_triples(Tg, R_I, force_phylogeny= force_phylogeny, update_lca= update_lca, inplace= inplace)
    elif algorithm=='prune_L':
        ret= prune_leaves(Tg, R_C, R_I, force_phylogeny= force_phylogeny, update_lca= update_lca, inplace= inplace, root= 1)
    else:
        raise ValueError(f'"{algorithm}" is not a valid algorithm for tree edition')

    return ret, len(R_I)

def prune_leaves(T, R_C, R_I, w_r= None, force_phylogeny= True, update_lca= True, inplace= True, root= 1):

    if not inplace:
        T= T.copy()
    if w_r==None:
        w_r= {r:1 for r in chain(R_C,R_I)}
    # Invert wheight of consistent triples
    w_r.update({rg:-w_r[rg] for rg in R_C})
    # Function for leaf prune-weight
    out_w= lambda x,G: sum((w_r[y] for y in G[x]))
    #out_w= lambda x,G: 0
    in_w= lambda x,G: sum((w_r[y[0]] for y in G.in_edges(x)))
    #in_w= lambda x,G: 0
    prune_w= lambda x,G: (out_w(x, G), in_w(x, G), x)

    # Construct three-partite graph
    leafs= set(filter(lambda x: is_leaf(T,x) , T))
    G= nx.DiGraph()
    G.add_nodes_from(chain(leafs,R_C,R_I))
    G.add_edges_from(((x,ri)
                      for x in leafs
                      for ri in R_I
                      if x in ri
                     ))
    G.add_edges_from(((rc,x)
                      for x in leafs
                      for rc in R_C
                      if x in rc
                     ))

    print("\n\n")
    LL= len(R_I)
    t0= time.time()
    print(len(leafs), len(R_I), len(R_C))

    # Choose subset of leafs
    prune= []
    while len(R_I)>0:
        leaf= _choose_best(G, leafs, prune_w)
        prune+= [leaf]
        t0= time.time()
        leafs= leafs-{leaf}
        t1= time.time()
        R_I= R_I-set(G[leaf])
        t2= time.time()
        R_C= R_C-set((y[0] for y in G.in_edges(leaf)))
        t3= time.time()
        G= nx.induced_subgraph(G, leafs.union(R_I).union(R_C) )
        t4= time.time()
        print('>>>>> ', t1-t0, t2-t1, t3-t2, t4-t3)

    t1= time.time()
    if LL==0:
        print('total per triple: ', '-----')
    else:
        print('total per triple: ', (t1-t0)/LL)

    # Obtain induced tree
    for x in prune:
        T.remove_node(x)
    # Compcat edges if necesary
    if force_phylogeny:
        nodes= list(filter(lambda x: len(T[x])==1,
                           dfs_postorder_nodes(T, root)
                          ))
        for x_node in nodes:
            x1= list(T[x_node])[0]
            compact_edge(T, x_node, x1, delete= 'upper', update_lca= False)

    # Update LCA
    if update_lca:
        set_sparce_matrix(T)
    return T

def _choose_best(G, leafs, prune_w):
    t0= time.time()
    df= [prune_w(x,G) for x in leafs]
    t1= time.time()
    best= sorted(df, reverse= True)[0][2]
    t2= time.time()
    print(">>> ", len(leafs), t1-t0, t2-t1)
    return best

def prune_triples(T, R, force_phylogeny= True, update_lca= True, inplace= True):
    if not inplace:
        T= T.copy()
    # Remove leafs present in the triples
    leaves= set(chain.from_iterable(R))
    for x in leaves:
        T.remove_node(x)

    # Compcat edges if necesary
    if force_phylogeny:
        nodes= list(filter(lambda x: len(T[x])==1,
                           dfs_postorder_nodes(T, T.root)
                          ))
        for x_node in nodes:
            x1= list(T[x_node])[0]
            compact_edge(T, x_node, x1, delete= 'upper', update_lca= False)

    # Update LCA
    if update_lca:
        set_sparce_matrix(T)

    return T

def correct_tree_df(df, Ts, tree_col= 'tree',
                    root= 1,
                    label_attr= 'label',
                    species_attr= 'species',
                    event_attr= 'event',
                    algorithm= 'prune_L',
                    inplace= False
                   ):
    if not inplace:
        df= df.copy()
    # Prepare species triples
    for y in Ts:
        if len(Ts[y])>0:
            Ts.nodes[y][event_attr]= 'S'
    Ts_triples= set(get_triplets(Ts, event=event_attr, color=species_attr, root_event='S'))


    # Correct trees
    out= df[tree_col].progress_apply(lambda T: correct_tree(T, Ts_triples, root= root,
                                                            label_attr= label_attr,
                                                     species_attr= species_attr,
                                                     event_attr= event_attr,
                                                     algorithm= algorithm,
                                                     force_phylogeny= True,
                                                     update_lca= True, 
                                                     inplace= True,
                ))
    df[tree_col]= out.str[0]
    df['edited']= out.str[1]
    return df

# Standalone usage
##################

if __name__ == "__main__":
    import pandas as pd

    import argparse
    parser = argparse.ArgumentParser(prog= 'revolutionhtl.tree_correction',
                                     description='Correction for gene tree with respect to a species tree',
                                     usage='python -m revolutionhtl.tree_correction <arguments>',
                                     formatter_class=argparse.MetavarTypeHelpFormatter,
                                    )

    # Arguments
    ###########

    # Input data
    # ..........

    parser.add_argument('gene_trees',
                        help= 'A .tsv file containing gene trees in the column specified by "-tree_column" in nhxx format',
                        type= str,
                       )

    parser.add_argument('species_tree',
                        help= '.nhxx file containing a species tree.',
                        type= str,
                       )

    # Parameters
    # ..........
    parser.add_argument('-alg', '--algorithm',
                        help= 'Algorithm for tree correction (default: prune_L).',
                        type= str,
                        choices= ['prune_L', 'prune_R'],
                        default= 'prune_L',
                       )

    # Format parameters
    # .................

    parser.add_argument('-T', '--tree_column',
                        help= 'Column containing trees in nhxx format at the "gene_trees" file. (default: tree).',
                        type= str,
                        default= 'tree'
                       )

    parser.add_argument('-o', '--output_prefix',
                        help= 'Prefix used for output files (default "tl_project").',
                        type= str,
                        default= 'tl_project',
                       )

    parser.add_argument('-T_attr', '--T_attr_sep',
                        help= 'String used to separate attributes in the gene trees. (default: ";").',
                        type= str,
                        default= ';',
                       )

    parser.add_argument('-S_attr', '--S_attr_sep',
                        help= 'String used to separate attributes in the species tree. (default: ";").',
                        type= str,
                        default= ';',
                       )

    args= parser.parse_args()

    # Perform edition
    #################

    print('\n---------------------------')
    print('\nREvolutionH-tl tree edition')
    print('---------------------------\n')

    print('Reading gene trees...')
    gTrees= pd.read_csv(args.gene_trees, sep= '\t')
    gTrees[args.tree_column]= gTrees[args.tree_column].progress_apply(
        lambda x: read_nhxx(x, name_attr= 'accession', attr_sep= args.T_attr_sep))

    print('Reading species tree...')
    with open(args.species_tree) as F:
        sTree= read_nhxx(''.join( F.read().strip().split('\n') ),
                         name_attr= 'species',
                         attr_sep= args.S_attr_sep
                        )

    print('Editing trees...')
    gTrees= correct_tree_df(gTrees, sTree, tree_col= args.tree_column,
                    root= 1,
                            label_attr= 'accession',
                    species_attr= 'species',
                    event_attr= 'accession',
                    algorithm= args.algorithm,
                    inplace= False
                   )

    print('Writting corrected trees...')
    opath= f'{args.output_prefix}.corrected_trees.tsv'
    gTrees.to_csv(opath, sep= '\t', index= False)
    print(f'Successfully written to {opath}')



from . import nxTree as nxt
import networkx as nx
from itertools import chain, product, combinations

def get_triplets(tree, event='event', color= 'color', root_event= 'S', loss_leafs= 'X'):
    """
    return a tuple (a,b,c), where a and b are the ingroup
    and c is the outgroup.
    """
    I= {} # Dictionary for induced leafs
    for x in nx.dfs_postorder_nodes(tree):
        if nxt.is_leaf(tree, x):
            if tree.nodes[x][color] == loss_leafs:
                I[x]= {  }
            else:
                I[x]= { tree.nodes[x][color] }
        else:
            if x!=tree.root:
                I[x]= set( chain.from_iterable((I[x1] for x1 in tree[x])) )
            if tree.nodes[x][event] == root_event:
                for triple in _get_triples_from_root(tree, x, I):
                    yield triple

def _get_triples_from_root(tree, node, I):
    for x0, x1 in combinations(tree[node], 2):
        for triple in chain(_get_triplets_from_groups(tree, x0, x1, I),
                            _get_triplets_from_groups(tree, x1, x0, I)):
            yield triple

def _get_triplets_from_groups(tree, x_out, x_in, I):
    P= combinations( I[x_in], 2 )
    for (a,(b,c)) in product( I[x_out], P ):
        if len({a,b,c})==3:
            yield tuple(sorted((b,c)))+(a,)



from .nxTree import is_leaf

import networkx as nx


def separe_loss_sons(T, x, label_attr= 'accession', loss_symbol= 'X'):
    L, Lx= [], []
    for x1 in T[x]:
        if T.nodes[x1][label_attr] == loss_symbol:
            Lx+= [x1]
        else:
            L+= [x1]
    return L, Lx

def prune_losses(T, label_attr= 'accession', loss_symbol= 'X'):
    # Init output tree
    Tp= nx.DiGraph()
    Tp.add_node(0, **T.nodes[0])
    Tp.root= 0
    # Add root to de queue and the dad of mu(root)
    Q= [(1, 0)]
    # Init counter of new nodes
    idx= 1

    while len(Q)>0:
        x, d= Q.pop(0)
        if is_leaf(T, x):
            Tp.add_node(idx, **T.nodes[x])
            Tp.add_edge(d, idx)
            idx+= 1
        else:
            L, _= separe_loss_sons(T, x, label_attr, loss_symbol)
            if len(L)>1:
                Tp.add_node(idx, **T.nodes[x])
                Tp.add_edge(d, idx)
                pDad= idx
                idx+= 1
            else:
                pDad= d
            Q+= [(x1, pDad) for x1 in L]
    return Tp



from .triplets import get_triplets
from .nhxx_tools import tralda_to_nxTree

from tralda.supertree.Build import greedy_BUILD
from bmgedit.Build import Build2
from tralda.datastructures import Tree
from itertools import chain
from collections import Counter

class WrongHeuristicError(Exception):
    """Raised when a input parameter specifies an heuristic not avaliable for an algorithm"""

def from_gene_forest(gTrees, method= 'louvain_weight', numb_repeats= 3, doble_build= True, binary= False):
    """
    Obtain triples rooted at symbol 'S' from a Series of gene trees.
    Then searchs for an approximated maximum set of consistent using an heuristic,
    and returns a species S_1 tree that explains such a set.

    "doble build" means that instead of returning the species tree S_1,
    triples will be retrivered from this tree and then run build in this
    set of triples, obtaining another tree S_2 which may be less resolved
    than S_1 see [REFERENCE].

    Input:
    - gTrees : pandas.Series of nxTree
               Each element is a gene tree where the node attributes are:
               - 'event' for gene name in leafs and symbol in inner nodes
               - 'color' for the species in the leafs
    - method : str
               Meta parameter of the heuristics, specifies a comunity-detection
               algorithm. See more in [REFERENCE]
    - numb_repeats : int
                     Meta parameter of the heuristics, specifies how many time
                     run the algorithm for community detection.
    - doble_build :  bool
                     Run build algorithm twice aimed to obtain less-resolved tree. See [REFERENCE]
    """
    weights= _count_triples(gTrees,
                           event='accession',
                           color= 'species',
                           root_event= 'S')
    triples= set(weights)
    leafs= set(chain.from_iterable(triples))
    
    T= build(triples,
             leafs,
             method= method,
             numb_repeats= numb_repeats,
             doble_build= doble_build,
             name_attr= 'species',
             weights= weights,
             binary= binary,
            )

    return T

def build(triples,
          leafs,
          method= 'louvain_weight',
          numb_repeats= 3,
          doble_build= True,
          name_attr= 'accession',
          weights= None,
          binary= False,
         ):
    if weights==None:
        weights= {x:1 for x in triples}

    if binary:
        binary= 'b'

    if method == 'naive': # add triples one by one and checks consistency via BUILD
        tree= greedy_BUILD(triples ,leafs)

    elif method in ('louvain','mincut'):
        build = Build2(triples, leafs,
                       allow_inconsistency=True,
                       part_method=method,
                       greedy_repeats=numb_repeats,
                       triple_weights= weights, #<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<is it used?
                       binarize= binary
                      )
        tree = build.build_tree() # use recursive Aho-algorithm with 1 greedy repeat

    elif method == 'louvain_weight':
        build = Build2(triples, leafs,
                       allow_inconsistency=True,
                       part_method='louvain',
                       greedy_repeats=numb_repeats,
                       weighted_mincut=True,
                       triple_weights= weights,
                       binarize= binary
                      )
        tree = build.build_tree() # use recursive Aho-algorithm with 1 greedy repeat
    else:
        raise WrongHeuristicError(f'The method "{method}" is not implemented. Avaliable: naive, louvain, mincut, louvain_weight')

    # Compute second tree
    if doble_build:
        common_triples=[tuple(x.label for x in triple) for triple in Tree.get_triples(tree)]
        build = Build2(common_triples, leafs,
                       allow_inconsistency=False)
        tree = build.build_tree() # use recursive Aho-algorithm

    # Transform to nxTree
    tree= tralda_to_nxTree(tree, name_attr= name_attr)

    return tree

def _count_triples(tree_list, **kwargs):
    triples= chain.from_iterable((get_triplets(T, **kwargs)
                                  for T in tree_list
                                 ))
    weights= Counter(triples)
    return weights
